from enum import Enum


class CordraOperation(Enum):
    GET_DESIGN = '20.DOIP/Op.GetDesign'
    GET_INIT_DATA = '20.DOIP/Op.GetInitData'


class DOIPOperation(Enum):
    HELLO = '0.DOIP/Op.Hello'
    CREATE = '0.DOIP/Op.Create'
    RETRIEVE = '0.DOIP/Op.Retrieve'
    UPDATE = '0.DOIP/Op.Update'
    DELETE = '0.DOIP/Op.Delete'
    SEARCH = '0.DOIP/Op.Search'
    LIST_OPERATION = '0.DOIP/Op.ListOperations'


class ResponseStatus(Enum):
    SUCCESS = '0.DOIP/Status.001'
    INVALID = '0.DOIP/Status.101'
    UNAUTHENTICATED = '0.DOIP/Status.102'
    UNAUTHORIZED = '0.DOIP/Status.103'
    UNKNOWN_DO = '0.DOIP/Status.104'
    DUPLICATED_PID = '0.DOIP/Status.105'
    UNKNOWN_OPERATION = '0.DOIP/Status.200'
    UNKNOWN_ERROR = '0.DOIP/Status.500'


class TypeIdentifier(Enum):
    FDO_PROFILE_REF = '21.T11969/bcc54a2a9ab5bf2a8f2c'
    FDO_TYPE_REF = '21.T11969/2bb5fec05c00bb89793e'
    FDO_RIGHTS_REF = '21.T11969/90fa2a1e224ae3e54139'
    FDO_GENRE_REF = '21.T11969/66d0e45fb2d2430ab967'
    FDO_DATA_REFS = '21.T11969/867134e94b3ec5afc6fe'
    FDO_MD_REFS = '21.T11969/a02253b264a9f2f1cf9a'


class ValidationSchemas(Enum):
    CREATE_FDO_DOIPY_INPUT = '21.T11969/6fe44ad5cba7c7c28c56'
    DATA_AND_METADATA = '21.T11969/27d8ca0c2585c0dafa39'


class DOType(Enum):
    DO = 'Document'
    FDO = 'FDO'


class DomainName(Enum):
    TYPE_API_SCHEMAS = 'https://typeapi.lab.pidconsortium.net/v1/types/schema'
    TYPE_API_OBJECTS = 'https://typeregistry.lab.pidconsortium.net/objects'
    TYPE_REGISTRY_OBJECTS = 'https://typeregistry.lab.pidconsortium.net/objects'



import json
from typing import Annotated

import typer
from jsonschema import ValidationError

from doipy import hello, list_operations, create_fdo, search, get_design, get_init_data, retrieve
from doipy.validation_utils import general_validation, get_specific_schema
from doipy.constants import ResponseStatus
from doipy.exceptions import AuthenticationException, InvalidRequestException, OperationNotSupportedException
from doipy.models import FdoInput
from doipy.server import server

app = typer.Typer()


@app.command(name='hello')
def hello_command(service: Annotated[str, typer.Argument(help='the PID identifying a data service in the service '
                                                              'registry')]):
    """
    Implements 0.DOIP/Op.Hello: An operation to allow a client to get information about the DOIP service.
    """
    response = hello(service)
    print(json.dumps(response, indent=2))


@app.command(name='list_operations')
def list_operations_command(service: Annotated[str, typer.Argument(help='the PID identifying a data service in the '
                                                                        'service registry')]):
    """
    Implements 0.DOIP/Op.ListOperations: An operation to request the list of operations that can be invoked on the
    target DO.
    """
    response = list_operations(service)
    print(json.dumps(response, indent=2))


'''
@app.command(name='create')
def create_command(
        service: Annotated[str, typer.Argument(help='the PID identifying a data service in the service registry')],
        input_file: Annotated[typer.FileText, typer.Argument(help='A file containing a JSON which follows a specific '
                                                                  'JSON schema. The file contains a bit-sequence and'
                                                                  'corresponding metadata which together form a DO.')],
        client_id: Annotated[str, typer.Option(help='The identifier of the user in Cordra')] = None,
        username: Annotated[str, typer.Option(help='Username of the user in Cordra')] = None,
        token: Annotated[str, typer.Option(help='Token generated in Cordra')] = None):

    """Implements 0.DOIP/Op.Create: An operation to create a digital object (containing at most one data bit-sequence)
    within the DOIP service. The target of a creation operation is the DOIP service itself."""

    # read the input
    input_json = json.load(input_file)

    # validate the input against the JSON input schema and define data structure for DO input
    try:
        do_input = DoInput.parse(input_json)
    except ValidationError as error:
        print(str(error))
        raise typer.Exit()

    # create the DO
    # if token, username and client_id are provided, authentication is successful if and only if token is correct
    # if username and client_id are both provided, authentication is successful if and only if client_id is correct
    try:
        if not token and (client_id or username):
            password = typer.prompt('Password', hide_input=True)
            response = create(service, DOType.DO.value, do_input.file, do_input.md, password, client_id, username,
                              token)
        else:
            response = create(service, do_type=DOType.DO.value, bitsq=do_input.file, metadata=do_input.md,
                              client_id=client_id, username=username, token=token)

    # in case that the DOIP response status is not success, raise an exception and terminate the program
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()
    except InvalidRequestException as error:
        details = error.args[0][0]
        print(f'{details["status"]} {ResponseStatus(details["status"]).name}')
        print(details['output']['message'])
        raise typer.Exit()

    print(json.dumps(response, indent=2))
'''


@app.command(name='create_fdo')
def create_fdo_command(
        input_file: Annotated[typer.FileText, typer.Argument(help='A file containing a JSON which follows a specific '
                                                                  'JSON schema. The file contains data bit-sequences, '
                                                                  'metadata bit-sequences and the metadata that should '
                                                                  'be written into the corresponding PID records.')]):
    """Create a new FAIR Digital Object (FDO) from data and metadata files."""

    # read the user input
    user_input = json.load(input_file)

    # validate the user input against the general input schema
    try:
        general_validation(user_input)
    except ValidationError as error:
        print(str(error))
        raise typer.Exit()

    # get the schema which belongs to the inputs of the create operation for the chosen profile
    try:
        specific_schema = get_specific_schema(user_input)
    except OperationNotSupportedException as error:
        print(str(error))
        raise typer.Exit()

    # validate the user input against the input schema for the specific profile and create an instance of FdoInput class
    try:
        fdo_input = FdoInput.parse(user_input, specific_schema)
    except ValidationError as error:
        print(str(error))
        raise typer.Exit()

    # Create the FDO
    try:
        response = create_fdo(fdo_input)
    # in case that authentication fails, raise an exception
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()
    # in case that the DOIP response is not success, raise an exception and terminate the program
    except InvalidRequestException as error:
        details = error.args[0][0]
        print(f'{details["status"]} {ResponseStatus(details["status"]).name}')
        print(details['output']['message'])
        raise typer.Exit()

    print(json.dumps(response, indent=2))


@app.command(name='retrieve')
def retrieve_command(service: Annotated[str, typer.Argument(help='the PID identifying a data service in the service '
                                                                 'registry')],
                     do: Annotated[str, typer.Argument(help='the PID identifying a (F)DO to be retrieved')]):
    """
    Implements 0.DOIP/Op.Retrieve: An operation to allow a client to get information about an (F)DO at a service.
    """
    response = retrieve(service, do)
    print(json.dumps(response, indent=2))


@app.command(name='search')
def search_command(query: Annotated[str, typer.Argument(help='query')],
                   username: Annotated[str, typer.Option(help='username')] = None,
                   password: Annotated[str, typer.Option(help='password')] = None):
    """Implements 0.DOIP/Op.Search"""
    response = search(query, username, password)
    print(json.dumps(response, indent=2))


@app.command(name='get_design')
def get_design_command():
    """Implements 20.DOIP/Op.GetDesign (see https://www.cordra.org)"""
    response = get_design()
    print(json.dumps(response, indent=2))


@app.command(name='get_init_data')
def get_init_data_command():
    """Implements 20.DOIP/Op.GetInitData (see https://www.cordra.org)"""
    response = get_init_data()
    print(json.dumps(response, indent=2))


@app.command(name='doip_server')
def doip_server_command(hostname: Annotated[str, typer.Argument(help='hostname')],
                        port: Annotated[int, typer.Argument(help='port')],
                        privkey: Annotated[str, typer.Argument(help='privkey')],
                        cert: Annotated[str, typer.Argument(help='cert')]):
    """Starts a DOIP server"""
    server(hostname, port, privkey, cert)



import requests
from jsonschema import validate

from doipy.constants import DomainName, ValidationSchemas
from doipy.exceptions import OperationNotSupportedException


def general_validation(user_input):
    url = f'{DomainName.TYPE_API_SCHEMAS.value}/{ValidationSchemas.CREATE_FDO_DOIPY_INPUT.value}'
    validation_schema = requests.get(url).json()
    validate(instance=user_input, schema=validation_schema)


def get_specific_schema(user_input):

    # get the service information from the dtr
    url = f'{DomainName.TYPE_API_OBJECTS.value}/{user_input["FDO_Service_Ref"]}'
    service = requests.get(url).json()

    # get all operations that can be performed on the given service
    operation_pids = service['implementsOperations']

    # find the correct create_FDO operation (which corresponds to the profile)
    found_operation = None
    for operation_pid in operation_pids:
        url = f'{DomainName.TYPE_REGISTRY_OBJECTS.value}/{operation_pid}'
        operation = requests.get(url).json()
        # condition that the correct FDO operation was found
        if (operation['operationCategory'] == 'Create_FDO' and operation['relatedFdoProfiles'][0]
                == user_input['FDO_Profile_Ref']):
            found_operation = operation
            break
    if not found_operation:
        raise OperationNotSupportedException('The service does not support Create_FDO for this profile.')

    # return input schema from the DTR for the chosen create_FDO operation
    input_pid = found_operation['inputs']
    url = f'{DomainName.TYPE_API_OBJECTS.value}/{input_pid}'
    validation_schema = requests.get(url).json()
    return validation_schema



import json
import socket
import ssl

from rich import print

from doipy.socket_utils import send_server_message


def server(hostname: str, port: int, privkey: str, cert: str):
    context = ssl.create_default_context()

    with socket.create_connection((hostname, 443)) as sock:
        with context.wrap_socket(sock, server_hostname=hostname) as ssock:
            print(ssock.version())

    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    context.load_cert_chain(cert, privkey)

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0) as sock:
        sock.bind(('127.0.0.1', port))
        sock.listen(5)
        with context.wrap_socket(sock, server_side=True) as ssock:
            print('Wait for connection')
            while True:
                conn, addr = ssock.accept()
                print('Connected by', addr)
                request = json.loads(conn.recv(8192))
                print(request)
                # data2 = conn.recv(8192)
                if request['operationId'] == '0.DOIP/Op.Hello':
                    doip_hello(conn, port)
                elif request['operationId'] == '0.DOIP/Op.ListOperations':
                    doip_list_operations(conn)
                else:
                    not_implemented(conn, request['operationId'])
                conn.close()


def doip_list_operations(conn: object):
    message = {
        'status': '0.DOIP/Status.001',
        'output': [
            '0.DOIP/Op.Hello',
            '0.DOIP/Op.ListOperations',
        ]
    }
    send_server_message(message, conn)


def doip_hello(conn: object, port: int):
    message = {'status': '0.DOIP/Status.001',
               'output': {
                   'id': 'doip-server-1',
                   'type': '0.TYPE/DOIPServiceInfo',
                   'attributes': {
                       'ipAddress': '0.0.0.0',
                       'port': port,
                       'protocol': 'TCP',
                       'protocolVersion': '2.0'
                   }
               }
               }
    print('send reply')
    send_server_message(message, conn)


def doip_create():
    print('dummy')


def doip_update():
    print('dummy')


def doip_delete():
    print('dummy')


def doip_search():
    print('dummy')


def not_implemented(conn: object, operation_id: str):
    message = {operation_id: 'not supported'}
    send_server_message(message, conn)



import uuid
from pathlib import Path

from doipy.exceptions import InvalidRequestException
from doipy.constants import DOIPOperation, ResponseStatus
from doipy.socket_utils import create_socket, send_message, finalize_socket, finalize_segment, read_response, \
    get_settings


def hello(service: str):
    # get service settings
    target_id, ip, port = get_settings(service)
    # create request message
    message = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.HELLO.value
    }
    # send request and return response
    with create_socket(ip, port) as ssl_sock:
        send_message(message, ssl_sock)
        finalize_segment(ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def list_operations(service):
    # get service settings
    target_id, ip, port = get_settings(service)
    # create request message
    message = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.LIST_OPERATION.value
    }

    # send request and read response
    with create_socket(ip, port) as ssl_sock:
        send_message(message, ssl_sock)
        finalize_segment(ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def create(service: str, do_type: str, do_name: str, bitsq: Path, metadata: dict, authentication_message: dict):

    # get service settings
    target_id, ip, port = get_settings(service)

    # create first message
    message_1 = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.CREATE.value
    }
    message_1 = message_1 | authentication_message

    # create second message: DO of type document in Cordra for the file which is added
    message_2 = {
        'type': do_type,
        'attributes': {
            'content': {
                'id': '',
                'name': do_name
            }
        }
    }

    # add metadata to DO
    if metadata:
        message_2['attributes']['content'] = message_2['attributes']['content'] | metadata

    with create_socket(ip, port) as ssl_sock:

        send_message(message_1, ssl_sock)
        finalize_segment(ssl_sock)

        if bitsq:
            # add information on files to DO
            filename = bitsq.name
            my_uuid = str(uuid.uuid4())
            message_2['elements'] = [
                {
                    'id': my_uuid,
                    'type': 'text/plain',
                    'attributes': {
                        'filename': filename
                    }
                }
            ]
            # third message
            message_3 = {
                'id': my_uuid
            }
            send_message(message_2, ssl_sock)
            finalize_segment(ssl_sock)

            send_message(message_3, ssl_sock)
            finalize_segment(ssl_sock)

            # send content of files
            buffer_size = 1024
            with open(bitsq, 'rb') as f:
                while bytes_read := f.read(buffer_size):
                    ssl_sock.sendall(bytes_read)
                finalize_segment(ssl_sock)

        finalize_socket(ssl_sock)

        response = read_response(ssl_sock)
    if response[0]['status'] == ResponseStatus.SUCCESS.value:
        return response
    raise InvalidRequestException(response)


def update(service: str, client_id: str, password: str, do_type: str):
    # TODO fix message

    # get service settings
    target_id, ip, port = get_settings(service)

    with create_socket(ip, port) as ssl_sock:
        message = {
            'clientId': client_id,
            'targetId': target_id,
            'operationId': DOIPOperation.UPDATE.value,
            'authentication': {
                'password': password
            }
        }
        send_message(message, ssl_sock)
        string1 = f'https://cordra.testbed.pid.gwdg.de/objects/{target_id}?payload=file'
        string2 = f'https://cordra.testbed.pid.gwdg.de/objects/{target_id}'
        message = {
            'type': do_type,
            'attributes': {
                'content': {
                    'id': '',
                    'Payload': string1,
                    'Metadata': string2
                }
            }
        }
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def search(service: str, query: str = 'type:Document', username: str = None, password: str = None):
    # TODO fix message

    # get service settings
    target_id, ip, port = get_settings(service)

    message = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.SEARCH.value,
        'attributes': {
            'query': query
        }
    }
    if username and password:
        message['authentication'] = {
            'username': username,
            'password': password
        }

    with create_socket(ip, port) as ssl_sock:
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def retrieve(service: str, do: str):
    # get service settings
    _, ip, port = get_settings(service)
    # create message
    message = {
        'targetId': do,
        'operationId': DOIPOperation.RETRIEVE.value
    }
    # send request and return response
    with create_socket(ip, port) as ssl_sock:
        send_message(message, ssl_sock)
        finalize_segment(ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


# delete



__all__ = [
    'create',
    'list_operations',
    'hello',
    'search',
    'update',
    'create_fdo',
    'get_design',
    'get_init_data',
    'retrieve'
]

from doipy.actions.cordra import get_design, get_init_data
from doipy.actions.doip import create, list_operations, hello, search, update, retrieve
from doipy.actions.fdo import create_fdo



import json

import requests
from jsonschema import validate
from pydantic import BaseModel
from pathlib import Path

from doipy.constants import DomainName, ValidationSchemas


class DoInput(BaseModel):
    # optional
    data_bitsq: Path = None
    data_values: dict = None
    metadata_bitsq: Path = None
    metadata_values: dict = None

    @classmethod
    def parse(cls, user_input: dict) -> 'DoInput':
        # validate the input against the JSON input schema
        url = f'{DomainName.TYPE_API_SCHEMAS.value}/{ValidationSchemas.DATA_AND_METADATA.value}'
        input_schema = requests.get(url).json()

        validate(instance=user_input, schema=input_schema)

        do_input = cls()

        if 'data_bitsq' in user_input:
            data_bitsq = Path(user_input['data_bitsq'])
            do_input.data_bitsq = data_bitsq

        if 'metadata_bitsq' in user_input:
            metadata_bitsq = Path(user_input['metadata_bitsq'])
            do_input.metadata_bitsq = metadata_bitsq

        if 'data_values' in user_input:
            with open(user_input['data_values']) as f:
                data_values = json.load(f)
                do_input.data_values = data_values

        if 'metadata_values' in user_input:
            with open(user_input['metadata_values']) as f:
                metadata_values = json.load(f)
                do_input.metadata_values = metadata_values

        return do_input


class FdoInput(BaseModel):
    # mandatory
    fdo_service_ref: str
    fdo_profile_ref: str
    authentication: dict
    fdo_type_ref: str
    # optional
    fdo_rights_ref: str = None
    fdo_genre_ref: str = None
    data_and_metadata: list[DoInput] = None

    @classmethod
    def parse(cls, user_input: dict, input_schema: dict) -> 'FdoInput':

        # validate the input against the JSON input schema
        validate(instance=user_input, schema=input_schema)

        # construct the fdo_input object with the mandatory references
        # TODO: can this be formulated as part of the DTR? -> ask Hans
        fdo_input = cls(fdo_service_ref=user_input['FDO_Service_Ref'], fdo_profile_ref=user_input['FDO_Profile_Ref'],
                        fdo_type_ref=user_input['FDO_Type_Ref'], authentication=user_input['FDO_Authentication'])

        # define the optional references
        if 'FDO_Rights_Ref' in user_input:
            fdo_input.fdo_rights_ref = user_input['FDO_Rights_Ref']
        if 'FDO_Genre_Ref' in user_input:
            fdo_input.fdo_genre_ref = user_input['FDO_Genre_Ref']

        # define the data and metadata arrays which will be used to create the DOs
        if user_input['FDO_Data_and_Metadata']:
            fdo_input.data_and_metadata = []
            for item in user_input['FDO_Data_and_Metadata']:
                if item:
                    # create the structure for the data DO and for the metadata DO
                    do = DoInput.parse(item)
                    fdo_input.data_and_metadata.append(do)

        return fdo_input



from doipy.actions.doip import create
from doipy.constants import DOIPOperation, DOType, TypeIdentifier, ResponseStatus
from doipy.exceptions import InvalidRequestException, AuthenticationException
from doipy.models import FdoInput
from doipy.socket_utils import create_socket, send_message, finalize_segment, finalize_socket, read_response, \
    get_settings


def create_fdo(fdo_input: FdoInput):

    # get service settings
    target_id, ip, port = get_settings(fdo_input.fdo_service_ref)

    # create first message
    message_1 = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.CREATE.value,
    }

    # check that either a username, password or token is provided
    authentication = fdo_input.authentication
    if 'username' not in authentication and 'clientId' not in authentication and 'token' not in authentication:
        raise AuthenticationException('Provide token, username or client_id')

    # provide correct set of authentication credentials
    authentication_message = {}
    if 'token' in authentication:
        authentication_message['authentication'] = {
            'token': authentication['token']
            }
    elif 'clientId' in authentication:
        if 'password' in authentication:
            authentication_message['clientId'] = authentication['client_id']
            authentication_message['authentication'] = {
                'password': authentication['password']
            }
        else:
            raise AuthenticationException('Provide password')
    else:
        if 'password' in authentication:
            authentication_message['authentication'] = {
                'username': authentication['username'],
                'password': authentication['password']
            }
        else:
            raise AuthenticationException('Provide password')
    message_1 = message_1 | authentication_message

    # create second message
    message_2 = {
        # TODO: type should not be part of the DOIP message, but of the adaptor (?)
        'type': DOType.FDO.value,
        'attributes': {
            'content': {
                'id': '',
                'name': 'FAIR Digital Object',
                # FDO_Profile_Ref: mandatory
                TypeIdentifier.FDO_PROFILE_REF.value: fdo_input.fdo_profile_ref,
                # FDO_Type_Ref: mandatory
                TypeIdentifier.FDO_TYPE_REF.value: fdo_input.fdo_type_ref
            }
        }
    }
    # FDO_Rights_Ref: optional
    if fdo_input.fdo_rights_ref:
        message_2['attributes']['content'][TypeIdentifier.FDO_RIGHTS_REF.value] = fdo_input.fdo_rights_ref
    # FDO_Genre_Ref: optional
    if fdo_input.fdo_genre_ref:
        message_2['attributes']['content'][TypeIdentifier.FDO_GENRE_REF.value] = fdo_input.fdo_genre_ref

    # create the data and metadata DOs
    if fdo_input.data_and_metadata:
        data_refs = []
        metadata_refs = []
        for item in fdo_input.data_and_metadata:
            # create the data do
            if item.data_bitsq or item.data_values:
                response = create(fdo_input.fdo_service_ref, DOType.DO.value, 'Data-DO', item.data_bitsq,
                                  item.data_values, authentication_message)
                data_ref = response[0]['output']['id']
                data_refs.append(data_ref)
            # create the metadata do
            if item.metadata_bitsq or item.metadata_values:
                response = create(fdo_input.fdo_service_ref, DOType.DO.value, 'Metadata-DO',
                                  item.metadata_bitsq, item.metadata_values, authentication_message)
                metadata_ref = response[0]['output']['id']
                metadata_refs.append(metadata_ref)
        if data_refs:
            message_2['attributes']['content'][TypeIdentifier.FDO_DATA_REFS.value] = data_refs
        if metadata_refs:
            message_2['attributes']['content'][TypeIdentifier.FDO_MD_REFS.value] = metadata_refs

        # create socket
        with create_socket(ip, port) as ssl_sock:

            # create the FDO
            send_message(message_1, ssl_sock)
            finalize_segment(ssl_sock)

            send_message(message_2, ssl_sock)
            finalize_segment(ssl_sock)
            finalize_socket(ssl_sock)

            response = read_response(ssl_sock)

        if response[0]['status'] == ResponseStatus.SUCCESS.value:
            return response
        raise InvalidRequestException(response)



from doipy.constants import CordraOperation
from doipy.socket_utils import create_socket, send_message, finalize_segment, finalize_socket, read_response, \
    get_settings


# get service settings
target_id, ip, port = get_settings('21.T11969/01370800d56a0d897dc1')


def get_design():
    message = {
        'targetId': f'{target_id}',
        'operationId': CordraOperation.GET_DESIGN.value
    }
    with create_socket(ip,port) as ssl_sock:
        send_message(message, ssl_sock)
        finalize_segment(ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def get_init_data():
    message = {
        'targetId': f'{target_id}',
        'operationId': CordraOperation.GET_INIT_DATA.value
    }
    with create_socket(ip, port) as ssl_sock:
        send_message(message, ssl_sock)
        finalize_segment(ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response



from contextlib import contextmanager
import socket
import ssl
import json

import requests

cached_data = {}


def get_settings(service: str):
    # get the targetId, IP and port from the service registry
    url = f'https://typeregistry.lab.pidconsortium.net/objects/{service}'

    if service not in cached_data:
        r = requests.get(url)
        # Currently we have just one prefix available. Hence, we take the first entry in the list of target ids. At a
        # later point of implementation, one might choose from multiple prefixes.
        data = r.json()
        target_id = data['targetId'][0]
        ip = data['ip']
        port = data['port']

        cached_data[service] = {
            'target_id': target_id,
            'ip': ip,
            'port': port
        }
    return cached_data[service]['target_id'], cached_data[service]['ip'], cached_data[service]['port']


@contextmanager
def create_socket(ip: str, port: int):
    # create context object to configure socket
    # negotiate protocol between client and server
    ssl_settings = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    # do not match peer certificates hostname
    ssl_settings.check_hostname = False
    # do not try to verify other peer's certificates
    ssl_settings.verify_mode = ssl.CERT_NONE

    # create SSL socket
    # create a new socket given address family and socket type
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        # wrap sock and return an SSL socket (ssl_sock) that is tied to the context
        with ssl_settings.wrap_socket(sock) as ssl_sock:
            # connect to a remote socket at address
            ssl_sock.connect((ip, port))
            yield ssl_sock


def send_message(message, ssl_sock):
    """Send a message to the server via ssl_sock without segment separator."""
    data = json.dumps(message).encode(encoding='UTF-8', errors='ignore')
    ssl_sock.send(data)


def send_server_message(message, conn):
    data = (json.dumps(message) + '\n').encode(encoding='UTF-8', errors='ignore')
    conn.send(data)
    conn.send(b'\n#\n')
    conn.send(b'#\n')


def finalize_segment(ssl_sock):
    """Finalize the current segment with separator \n#\n"""
    ssl_sock.send(b'\n#\n')


def finalize_socket(ssl_sock):
    """Finalize the current message with separator #/n"""
    ssl_sock.send(b'#\n')


def read_response(ssl_sock):
    """Read and store the response from the server"""
    recv_message = ''
    resp_list = []
    byte_size = 1024

    while True:
        # read from buffer and save message
        received = ssl_sock.recv(byte_size)
        # end of segment
        if received == b'\n#\n':
            recv_message = json.loads(recv_message)
            resp_list.append(recv_message)
            received = ssl_sock.recv(byte_size)
            # end of response
            if received == b'#\n':
                return resp_list
            # read beginning of next segment
            recv_message = ''
            continue

        # concatenate messages
        recv_message += received.decode()
